@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # "screenkey" is distributed under GNU GPLv3+, WITHOUT ANY WARRANTY.
 # Copyright(c) 2010-2012: Pablo Seminario <pabluk@gmail.com>
 # Copyright(c) 2015-2020: wave++ "Yuri D'Elia" <wavexx@thregr.org>
@@ -7,10 +6,12 @@
 from . import *
 from .labelmanager import LabelManager
 
+from datetime import datetime
 import json
 import os
 import subprocess
 import numbers
+from tempfile import NamedTemporaryFile
 
 import gi
 gi.require_version('Gtk', '3.0')
@@ -19,10 +20,11 @@ gi.require_version('Pango', '1.0')
 from gi.repository import GLib
 GLib.threads_init()
 
-from gi.repository import Gtk, Gdk, Pango, GObject
+from gi.repository import Gtk, Gdk, GdkPixbuf, Pango, GObject
 import cairo
 
 
+# Gtk shortcuts
 START = Gtk.Align.START
 CENTER = Gtk.Align.CENTER
 END = Gtk.Align.END
@@ -36,6 +38,45 @@ VERTICAL = Gtk.Orientation.VERTICAL
 IF_VALID = Gtk.SpinButtonUpdatePolicy.IF_VALID
 
 
+BUTTONS_MIN_BLINK = 1/30        # Minimum persistence for any action (s)
+BUTTONS_REL_BRIGHT = 127        # Residual brightness after button release
+
+
+# SVG Data for mouse buttons
+BUTTONS_SVG = None
+
+def load_button_pixbufs(color):
+    global BUTTONS_SVG
+
+    if BUTTONS_SVG is None:
+        image_path = os.path.join(MODULE_DIR, 'images', 'mouse.svg')
+        with open(image_path) as svg_file:
+            BUTTONS_SVG = svg_file.readlines()
+
+    if not isinstance(color, str):
+        # Gdk.Color
+        color = 'rgb({}, {}, {})'.format(
+            round(color.red_float * 255),
+            round(color.green_float * 255),
+            round(color.blue_float * 255)
+        )
+    button_pixbufs = []
+    svg = NamedTemporaryFile(mode='w', suffix='.svg')
+    for line in BUTTONS_SVG[1:-1]:
+        svg.seek(0)
+        svg.truncate()
+        svg.writelines((
+            BUTTONS_SVG[0],
+            line.replace('#fff', color),
+            BUTTONS_SVG[-1],
+        ))
+        svg.flush()
+        os.fsync(svg.fileno())
+        button_pixbufs.append(GdkPixbuf.Pixbuf.new_from_file(svg.name))
+    svg.close()
+    return button_pixbufs
+
+
 def gi_module_available(module, version):
     try:
         gi.require_version(module, version)
@@ -78,7 +119,9 @@ class Screenkey(Gtk.Window):
                             'vis_space': True,
                             'geometry': None,
                             'screen': 0,
-                            'start_disabled': False})
+                            'start_disabled': False,
+                            'mouse': False,
+                            'button_hide_duration': 1})
         self.options = self.load_state()
         if self.options is None:
             self.options = defaults
@@ -98,14 +141,23 @@ class Screenkey(Gtk.Window):
         self.set_focus_on_map(False)
         self.set_app_paintable(True)
 
+        self.button_pixbufs = []
+        self.button_states = [None] * 11
+        self.img = Gtk.Image()
+        self.update_image_tag = None
+
+        self.box = Gtk.HBox(homogeneous=False)
+        self.box.show()
+        self.add(self.box)
+
         self.label = Gtk.Label()
         self.label.set_ellipsize(Pango.EllipsizeMode.START)
         self.label.set_justify(Gtk.Justification.CENTER)
         self.label.show()
-        self.add(self.label)
 
         self.font = Pango.FontDescription(self.options.font_desc)
         self.update_colors()
+        self.update_mouse_enabled()
 
         self.set_size_request(0, 0)
         self.set_gravity(Gdk.Gravity.CENTER)
@@ -113,7 +165,7 @@ class Screenkey(Gtk.Window):
         self.connect("draw", self.on_draw)
 
         scr = self.get_screen()
-        scr.connect("size-changed", self.on_configure)
+        scr.connect("size-changed", self.on_size_changed)
         scr.connect("monitors-changed", self.on_monitors_changed)
         self.set_active_monitor(self.options.screen)
 
@@ -121,6 +173,9 @@ class Screenkey(Gtk.Window):
         if visual is not None:
             self.set_visual(visual)
 
+        self.box.pack_start(self.img, expand=False, fill=True, padding=0)
+        self.box.pack_end(self.label, expand=True, fill=True, padding=0)
+
         self.labelmngr = None
         self.enabled = True
         self.on_change_mode()
@@ -152,10 +207,10 @@ class Screenkey(Gtk.Window):
         """Load stored options"""
         options = None
         try:
-            with open(self.STATE_FILE, 'r') as f:
+            with open(self.STATE_FILE) as f:
                 options = Options(json.load(f))
                 self.logger.debug("Options loaded.")
-        except IOError:
+        except OSError:
             self.logger.debug("file %s does not exists." % self.STATE_FILE)
         except ValueError:
             self.logger.debug("file %s is invalid." % self.STATE_FILE)
@@ -173,7 +228,7 @@ class Screenkey(Gtk.Window):
             with open(self.STATE_FILE, 'w') as f:
                 json.dump(options._store, f)
                 self.logger.debug("Options saved.")
-        except IOError:
+        except OSError:
             self.logger.debug("Cannot open %s." % self.STATE_FILE)
 
 
@@ -190,17 +245,83 @@ class Screenkey(Gtk.Window):
         self.set_active_monitor(self.monitor)
 
 
+    def update_mouse_enabled(self):
+        if self.options.mouse:
+            if not self.button_pixbufs:
+                self.button_pixbufs = load_button_pixbufs(
+                    Gdk.color_parse(self.options.font_color)
+                )
+            self.img.show()
+            self.update_image_tag = GLib.idle_add(self.update_image)
+        else:
+            self.img.hide()
+            if self.update_image_tag is not None:
+                GLib.source_remove(self.update_image_tag)
+                self.update_image_tag = None
+
+
+    def do_get_preferred_height(self):
+        return self.height
+
+
     def update_font(self):
-        _, window_height = self.get_size()
         text = self.label.get_text()
         lines = text.count('\n') + 1
-        self.font.set_absolute_size((50 * window_height // lines // 100) * 1000)
+        self.font.set_absolute_size((50 * self.height // lines // 100) * 1000)
+        self.label.set_padding(self.width // 100, 0)
         self.label.get_pango_context().set_font_description(self.font)
 
 
+    def update_image(self):
+        if not self.button_pixbufs:
+            self.update_image_tag = None
+            return False
+
+        pixbuf = self.button_pixbufs[0]
+        copied = False
+
+        for index, button_state in enumerate(self.button_states):
+            if button_state is None:
+                continue
+            delta_time = (datetime.now() - button_state.stamp).total_seconds()
+            if button_state.pressed or delta_time < BUTTONS_MIN_BLINK:
+                alpha = 255
+            elif self.options.button_hide_duration > 0:
+                hide_time = delta_time / self.options.button_hide_duration
+                alpha = int(BUTTONS_REL_BRIGHT * (1 - min(1, hide_time)))
+            else:
+                alpha = 0
+
+            if not copied:
+                pixbuf = pixbuf.copy()
+                copied = True
+            self.button_pixbufs[button_state.btn].composite(
+                pixbuf, 0, 0, pixbuf.get_width(), pixbuf.get_height(),
+                0, 0, 1, 1,
+                GdkPixbuf.InterpType.NEAREST, alpha)
+
+            if not button_state.pressed and \
+               delta_time >= self.options.button_hide_duration:
+                self.button_states[index] = None
+
+        scale = self.height / pixbuf.get_height()
+        if scale != 1:
+            width = int(pixbuf.get_width() * scale)
+            pixbuf = pixbuf.scale_simple(width, self.height, GdkPixbuf.InterpType.BILINEAR)
+        self.img.set_from_pixbuf(pixbuf)
+
+        if not copied:
+            self.update_image_tag = None
+            return False
+        return True
+
+
     def update_colors(self):
-        self.label.modify_fg(Gtk.StateFlags.NORMAL, Gdk.color_parse(self.options.font_color))
+        font_color = Gdk.color_parse(self.options.font_color)
+        self.label.modify_fg(Gtk.StateFlags.NORMAL, font_color)
         self.bg_color = Gdk.color_parse(self.options.bg_color)
+        if self.options.mouse and self.button_pixbufs:
+            self.button_pixbufs = load_button_pixbufs(font_color)
         self.queue_draw()
 
 
@@ -211,46 +332,51 @@ class Screenkey(Gtk.Window):
                            self.options.opacity)
         cr.set_operator(cairo.OPERATOR_SOURCE)
         cr.paint()
+        cr.set_operator(cairo.OPERATOR_OVER)
         return False
 
 
     def on_configure(self, *_):
-        window_x, window_y = self.get_position()
-        window_width, window_height = self.get_size()
-
         # set event mask for click-through
         self.input_shape_combine_region(cairo.Region(cairo.RectangleInt(0, 0, 0, 0)))
 
-        # set some proportional inner padding
-        self.label.set_padding(window_width // 100, 0)
 
+    def on_size_changed(self):
+        self.width, self.height = self.get_size()
         self.update_font()
+        self.update_image()
 
 
     def update_geometry(self, configure=False):
         geometry = self.get_screen().get_monitor_geometry(self.monitor)
+
         if self.options.geometry is not None:
+            # NOTE: this assume a single global scaling factor for all
+            # monitors which seems to be true for GTK3:
+            # https://wiki.gnome.org/HowDoI/HiDpi
+            scale = self.get_screen().get_monitor_scale_factor(self.monitor)
+
+            # scale the coordinate according to the selected monitor
+            # or convert from device to gdk pixel units
+            def coord_to_gdk(coord, extent):
+                if not isinstance(coord, numbers.Integral):
+                    return int(coord * extent)
+                else:
+                    return int(coord / scale)
+
             x, y, w, h = self.options.geometry
-            if not isinstance(x, numbers.Integral):
-                x = int(x * geometry.width)
-            if not isinstance(y, numbers.Integral):
-                y = int(y * geometry.height)
-            if not isinstance(w, numbers.Integral):
-                w = int(w * geometry.width)
-            if not isinstance(h, numbers.Integral):
-                h = int(h * geometry.height)
-            if x < 0:
-                x = geometry.width + x - w
-            if y < 0:
-                y = geometry.height + y - h
+            x = coord_to_gdk(x, geometry.width)
+            y = coord_to_gdk(y, geometry.height)
+            w = coord_to_gdk(w, geometry.width)
+            h = coord_to_gdk(h, geometry.height)
+            if x < 0: x = geometry.width + x - w
+            if y < 0: y = geometry.height + y - h
             area_geometry = [x, y, w, h]
         else:
             area_geometry = [geometry.x, geometry.y, geometry.width, geometry.height]
 
         if self.options.position == 'fixed':
-            self.move(*area_geometry[0:2])
-            self.resize(*area_geometry[2:4])
-            self.update_font()
+            self.move_resize(*area_geometry)
             return
 
         if self.options.font_size == 'large':
@@ -259,7 +385,6 @@ class Screenkey(Gtk.Window):
             window_height = 12 * area_geometry[3] // 100
         else:
             window_height = 8 * area_geometry[3] // 100
-        self.resize(area_geometry[2], window_height)
 
         if self.options.position == 'top':
             window_y = area_geometry[1] + area_geometry[3] // 10
@@ -267,8 +392,16 @@ class Screenkey(Gtk.Window):
             window_y = area_geometry[1] + area_geometry[3] // 2 - window_height // 2
         else:
             window_y = area_geometry[1] + area_geometry[3] * 9 // 10 - window_height
-        self.move(area_geometry[0], window_y)
-        self.update_font()
+
+        self.move_resize(area_geometry[0], window_y,
+                         area_geometry[2], window_height)
+
+
+    def move_resize(self, x, y, w, h):
+        self.width = w
+        self.height = h
+        self.move(x, y)
+        self.resize(w, h)
 
 
     def on_statusicon_popup(self, widget, button, timestamp, data=None):
@@ -277,10 +410,6 @@ class Screenkey(Gtk.Window):
             data.popup_at_pointer(None)
 
 
-    def show(self):
-        super(Screenkey, self).show()
-
-
     def on_labelmngr_error(self):
         msg = Gtk.MessageDialog(parent=self,
                                 type=Gtk.MessageType.ERROR,
@@ -296,6 +425,16 @@ class Screenkey(Gtk.Window):
         self.quit(exit_status=os.EX_SOFTWARE)
 
 
+    def timed_show(self):
+        if not self.get_property('visible'):
+            self.show()
+        if self.timer_hide is not None:
+            GObject.source_remove(self.timer_hide)
+            self.timer_hide = None
+        if self.options.timeout > 0 and not any(b and b.pressed for b in self.button_states):
+            self.timer_hide = GObject.timeout_add(self.options.timeout * 1000, self.on_timeout_main)
+
+
     def on_label_change(self, markup, synthetic):
         if markup is None:
             self.on_labelmngr_error()
@@ -306,18 +445,33 @@ class Screenkey(Gtk.Window):
         self.label.set_attributes(attr)
         self.update_font()
 
-        if not self.get_property('visible'):
-            self.show()
-        if self.timer_hide is not None:
-            GObject.source_remove(self.timer_hide)
-        if self.options.timeout > 0:
-            self.timer_hide = GObject.timeout_add(self.options.timeout * 1000, self.on_timeout_main)
+        self.timed_show()
         if self.timer_min is not None:
             GObject.source_remove(self.timer_min)
+            self.timer_min = None
         if not synthetic:
             self.timer_min = GObject.timeout_add(self.options.recent_thr * 2000, self.on_timeout_min)
 
 
+    def on_image_change(self, button_state):
+        if button_state:
+            btn = button_state.btn
+            # Don't do animation after stealth enable
+            if self.button_states[btn] is not None or button_state.pressed:
+                self.button_states[btn] = button_state
+                if self.options.mouse:
+                    if not self.update_image_tag:
+                        self.update_image_tag = GLib.idle_add(self.update_image)
+                    self.timed_show()
+        else:
+            # Reset all
+            self.button_states = [None for _ in self.button_states]
+            if self.options.mouse:
+                if not self.update_image_tag:
+                    self.update_image_tag = GLib.idle_add(self.update_image)
+                self.timed_show()
+
+
     def on_timeout_main(self):
         if not self.options.persist:
             self.hide()
@@ -337,7 +491,9 @@ class Screenkey(Gtk.Window):
         self.logger.debug("Restarting LabelManager.")
         if self.labelmngr:
             self.labelmngr.stop()
-        self.labelmngr = LabelManager(self.on_label_change, logger=self.logger,
+        self.labelmngr = LabelManager(self.on_label_change,
+                                      self.on_image_change,
+                                      logger=self.logger,
                                       key_mode=self.options.key_mode,
                                       bak_mode=self.options.bak_mode,
                                       mods_mode=self.options.mods_mode,
@@ -522,6 +678,15 @@ class Screenkey(Gtk.Window):
             self.font = widget.props.font_desc
             self.update_font()
 
+        def on_cbox_mouse_changed(widget, data=None):
+            self.options.mouse = widget.get_active()
+            self.logger.debug("Mouse changed: %s." % self.options.mouse)
+            self.update_mouse_enabled()
+
+        def on_sb_mouse_duration_changed(widget, data=None):
+            self.options.button_hide_duration = widget.get_value()
+            self.logger.debug("Button hide duration value changed: %f." % self.options.button_hide_duration)
+
         frm_time = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Time"),
                                                     use_markup=True),
                              border_width=4,
@@ -740,6 +905,36 @@ class Screenkey(Gtk.Window):
         grid_color.attach_next_to(adj_scale, lbl_opacity, RIGHT, 1, 1)
         frm_color.add(grid_color)
 
+        frm_mouse = Gtk.Frame(label_widget=Gtk.Label("<b>%s</b>" % _("Mouse"),
+                                                    use_markup=True),
+                             border_width=4,
+                             shadow_type=Gtk.ShadowType.NONE,
+                             margin=6, hexpand=True)
+        vbox_mouse = Gtk.VBox(spacing=6)
+
+        chk_mouse = Gtk.CheckButton(_("Show Mouse"))
+        chk_mouse.connect("toggled", on_cbox_mouse_changed)
+        chk_mouse.set_active(self.options.mouse)
+        vbox_mouse.pack_start(chk_mouse, expand=False, fill=True, padding=0)
+
+        hbox_mouse = Gtk.HBox()
+        lbl_mouse1 = Gtk.Label(_("Hide duration"))
+        lbl_mouse2 = Gtk.Label(_("seconds"))
+        sb_mouse = Gtk.SpinButton(digits=1)
+        sb_mouse.set_increments(0.5, 1.0)
+        sb_mouse.set_range(0.0, 2.0)
+        sb_mouse.set_numeric(True)
+        sb_mouse.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)
+        sb_mouse.set_value(self.options.button_hide_duration)
+        sb_mouse.connect("value-changed", on_sb_mouse_duration_changed)
+        hbox_mouse.pack_start(lbl_mouse1, expand=False, fill=False, padding=6)
+        hbox_mouse.pack_start(sb_mouse, expand=False, fill=False, padding=4)
+        hbox_mouse.pack_start(lbl_mouse2, expand=False, fill=False, padding=4)
+        vbox_mouse.pack_start(hbox_mouse, expand=False, fill=False, padding=6)
+
+        frm_mouse.add(vbox_mouse)
+        frm_mouse.show_all()
+
         hbox_main = Gtk.Grid(column_homogeneous=True)
         vbox_main = Gtk.Grid(orientation=VERTICAL)
         vbox_main.add(frm_time)
@@ -749,6 +944,7 @@ class Screenkey(Gtk.Window):
         vbox_main = Gtk.Grid(orientation=VERTICAL)
         vbox_main.add(frm_kbd)
         vbox_main.add(frm_color)
+        vbox_main.add(frm_mouse)
         hbox_main.add(vbox_main)
 
         box = prefs.get_content_area()
